package pro.litvinovg.docsettings.document;

import static pro.litvinovg.docsettings.settings.StorageUtils.getDocFromString;
import static pro.litvinovg.docsettings.settings.StorageUtils.getSettingsElement;

import java.awt.EventQueue;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.nio.charset.StandardCharsets;
import java.nio.file.FileSystem;
import java.nio.file.FileSystems;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.StandardOpenOption;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.apache.tika.Tika;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import com.sun.star.beans.PropertyValue;
import com.sun.star.beans.XPropertyContainer;
import com.sun.star.beans.XPropertySet;
import com.sun.star.beans.XPropertySetInfo;
import com.sun.star.chart.TimeUnit;
import com.sun.star.document.XDocumentProperties;
import com.sun.star.document.XDocumentPropertiesSupplier;
import com.sun.star.frame.XComponentLoader;
import com.sun.star.frame.XDesktop;
import com.sun.star.frame.XDispatchHelper;
import com.sun.star.frame.XDispatchProvider;
import com.sun.star.frame.XModel;
import com.sun.star.frame.XStorable;
import com.sun.star.lang.XComponent;
import com.sun.star.lang.XMultiComponentFactory;
import com.sun.star.lang.XServiceInfo;
import com.sun.star.text.XTextDocument;
import com.sun.star.uno.UnoRuntime;
import com.sun.star.uno.XComponentContext;
import com.sun.star.util.XModifiable;

import pro.litvinovg.docsettings.DialogHelper;
import pro.litvinovg.docsettings.Localizer;
import pro.litvinovg.docsettings.gui.MainWindow;
import pro.litvinovg.docsettings.settings.Backups;
import pro.litvinovg.docsettings.settings.ODFSetting;
import pro.litvinovg.docsettings.settings.SettingFactory;
import pro.litvinovg.docsettings.settings.Settings;
import pro.litvinovg.docsettings.settings.SettingsComparator;

public class ODFDocument {


	private static final Logger logger = LoggerFactory.getLogger(ODFDocument.class);

	private XComponentContext context = null;
	private XDesktop xDesktop = null;
	private XMultiComponentFactory multiComponentFactory = null;
	private XComponent currentDocument = null;
	private String fileName = "";
  public static final String SETTINGS = "settings.xml";
  public static final String META = "meta.xml";
  private Settings currentSettings = null;

	private XStorable storable = null;

	private XComponentLoader loader = null;

	private String url = null;

	public Settings getCurrentSettings() {
		if (currentSettings.getSize() == 0) {
			readSettings();
		}
		return currentSettings;
	}

	private String getSettingsContent() {
		return readInternalFile(SETTINGS);
	}

	private void setSettingsContent(String settingsContent) {
		writeInternalFile(SETTINGS, settingsContent);
	}

	private String getMetaContent() {
		return readInternalFile(META);
	}
	
	/*
	 * private void setMetaContent(String newMetaContent) { writeInternalFile(META,
	 * newMetaContent); }
	 */
	
	public ODFDocument(XComponentContext componentContext) throws Exception {
		this.currentSettings = new Settings();
		if (componentContext != null) {
			context = componentContext;
			multiComponentFactory = context.getServiceManager();

			Object oDesktop = multiComponentFactory.createInstanceWithContext("com.sun.star.frame.Desktop", context);
			xDesktop = UnoRuntime.queryInterface(XDesktop.class, oDesktop);
			currentDocument = xDesktop.getCurrentComponent();
			if (currentDocument == null) {
				logger.debug("xDesktop.getCurrentComponent() returned null");
				throw new Exception(Localizer.get("error_document_is_null"));
			}
			XModel model = UnoRuntime.queryInterface(XModel.class, currentDocument);
			if (!isSupported(currentDocument, "com.sun.star.text.TextDocument")) {
				throw new Exception(Localizer.get("error_document_type_is_not_supported"));
			}
			loader = UnoRuntime.queryInterface(XComponentLoader.class, xDesktop);
			XTextDocument textDoc = UnoRuntime.queryInterface(XTextDocument.class, currentDocument);
			if (textDoc == null) {
				throw new Exception("Error. TextDoc is null.");
			}
			storable = UnoRuntime.queryInterface(XStorable.class, textDoc);

			if (storable == null || storable.isReadonly()) {
				throw new Exception(Localizer.get("error_document_is_ro"));
			}
			if (!storable.hasLocation() || storable.getLocation().isEmpty()) {
				throw new Exception(Localizer.get("error_document_is_not_saved"));
			}
			
			url = model.getURL();
			URI uri = new URI(url);
			File file = new File(uri);
			if (!file.exists()) {
				throw new Exception(Localizer.get("error_document_is_null"));
			} 
			if (!file.canRead()) {
				throw new Exception(Localizer.get("error_document_is_ro"));
			}
			if (!"application/vnd.oasis.opendocument.text".equals(getMimeType(file))) {
				throw new Exception(Localizer.get("error_document_type_is_not_supported"));
			}
			fileName = file.getAbsolutePath();
		}
	}
	
	private String getMimeType(File file) {
		Tika tika = new Tika();
		InputStream is;
		String mime = "";
		try {
			is = new FileInputStream(file);
			mime = tika.detect(is);
		} catch (IOException e) {
			logger.debug("Ops", e);
		}
		return mime;
	}
	
	public void restoreSettings(Settings settings) {
		close();
		restoreFrom(settings);
		reopen();
	}
	
	public XPropertyContainer getUserDefinedProps(){
    XDocumentPropertiesSupplier docPropertiesSupplier = UnoRuntime.queryInterface(XDocumentPropertiesSupplier.class, currentDocument);
    XDocumentProperties docProperties = docPropertiesSupplier.getDocumentProperties();
    XPropertyContainer userDefinedProperties = docProperties.getUserDefinedProperties();
    return userDefinedProperties;
  } 
		
	public void readSettings() {
		flushSettings();
		try {
			Node configSettings = getSettingsElement(getDocFromString(getSettingsContent()));
			NodeList settingNodes = configSettings.getChildNodes();
			for (int i = 0; i < settingNodes.getLength(); i++) {
				Node settingNode = settingNodes.item(i);
				ODFSetting setting = SettingFactory.create(settingNode);
				currentSettings.add(setting);
			}
		} catch (Exception e) {
			logger.debug(e.getMessage());
			logger.debug("Ops!",e);
		}
	}
	
	private void flushSettings() {
		currentSettings.clear();
	}

	public void restoreFrom(Settings settings) {
			String newSettingsContent;
			try {
				newSettingsContent = settings.toSettingsContent(getSettingsContent());
				writeInternalFile(SETTINGS, newSettingsContent);

			} catch (Exception e) {
				logger.debug("Restore failed", e);
			}
	}
	
	public void readBackups(Backups backups) {
		backups.readBackupsFromMetadata(getUserDefinedProps());
	}
	
	public boolean hasBackups(Backups backups) {
		return !backups.isEmpty();
	}
	
	public void writeBackups(Backups backups) {
		backups.writeBackupsToMetadata(getUserDefinedProps());
	}
	
	public String getFileName() {
		return fileName;
	}
	
	public void printText(String text) {
		DialogHelper.showErrorMessage(context, null, text);
	}
	
	public void printFileName() {
		DialogHelper.showErrorMessage(context, null, fileName);
	}

	private String readInternalFile(String internalFile) {
		Map<String, String> env = new HashMap<>();
		String result = "";
		FileSystem zipfs = null;
		env.put("create", "true");
		Path path = Paths.get(fileName);
		URI uri = URI.create("jar:" + path.toUri());
		try {
			zipfs = FileSystems.newFileSystem(uri, env);
			Path settingsPath = zipfs.getPath(internalFile);
			byte[] encoded = Files.readAllBytes(settingsPath);
			result = new String(encoded, "UTF-8");
		} catch (IOException e) {
			logger.debug(e.getMessage());
			logger.debug("Ops!",e);
			e.printStackTrace();
		} finally {
			try {
				zipfs.close();
			} catch (IOException e) {
				logger.debug("Ops", e);
				e.printStackTrace();
			}
		}
		return result;
	}
	
	private void writeInternalFile(String internalFile, String contents) {
		Map<String, String> env = new HashMap<>();
		BufferedWriter writer = null;
		FileSystem zipfs = null;
		env.put("create", "true");
		Path path = Paths.get(fileName);
		URI uri = URI.create("jar:" + path.toUri());
		try {
			zipfs = FileSystems.newFileSystem(uri, env);
			Path internalPath = zipfs.getPath(internalFile);
			writer = Files.newBufferedWriter(internalPath, StandardCharsets.UTF_8, StandardOpenOption.WRITE);
			writer.write(contents);
		} catch (IOException e) {
			logger.debug(e.getMessage());
			logger.debug("Ops!",e);
		} finally {
			try {
				writer.close();
				zipfs.close();
			} catch (IOException e) {
				logger.debug(e.getLocalizedMessage());
				logger.debug("Ops!",e);
			}
		}
	}
	
	
	public void reopen(){
		try {
			XComponent doc = loader.loadComponentFromURL(url, "_default", 0, new PropertyValue[0]);
		} catch (Exception e) {
			System.out.println("Failed to open the document");
		}
		;
	} 
	
	public void save() {
		if (context != null) {
			executedispatch("Save", new PropertyValue[0]);	
		} else {
			logger.debug("Can't save document. No context.");
		}
	}
	
	public void close() {
		if (context != null) {
			executedispatch("CloseDoc", new PropertyValue[0]);	
		} else {
			logger.debug("Can't close document. No context.");
		}
	}
	
	public void reload() {
		if (context != null) {
			executedispatch("Reload", new PropertyValue[0]);	
		} else {
			logger.debug("Can't reload document. No context.");
		}
	}
	
	private boolean isSupported(Object object, String service) {
		XServiceInfo info = UnoRuntime.queryInterface(XServiceInfo.class, object);
		if (info == null) {
			return false;
		}
		return info.supportsService(service);
	}
	
	private boolean executedispatch( String cmd, PropertyValue[] props) {
		XDispatchHelper helper = createInstance(XDispatchHelper.class, "com.sun.star.frame.DispatchHelper");
		if (helper == null) {
			logger.debug("Could not create dispatch helper for command " + cmd);
			return false;
		}
		try {
		  Object oDesktop = context.getServiceManager().createInstanceWithContext("com.sun.star.frame.Desktop", context);
		  XDesktop xDesktop = UnoRuntime.queryInterface(XDesktop.class, oDesktop);
			XDispatchProvider provider = UnoRuntime.queryInterface(XDispatchProvider.class, xDesktop.getCurrentFrame());
			helper.executeDispatch(provider, ".uno:" + cmd, "", 0, props);
			return true;
		} catch (java.lang.Exception e) {
			logger.debug("Could not dispatch '" + cmd + "':\n  " + e);
		}
		return false;
	}

	private <T> T createInstance(Class<T> aType, String serviceName) {
		if ((context == null)) {
			logger.debug("No libreoffice context. Cant create instance of " + aType.toString());
			return null;
		}
		T interfaceObj = null;
		try {
			Object o = context.getServiceManager().createInstanceWithContext(serviceName, context);
			interfaceObj = UnoRuntime.queryInterface(aType, o);
		} catch (Exception e) {
			logger.debug("Couldn't create interface for '" + serviceName + "': " + e);
		}
		return interfaceObj;
	}

	public boolean compareWithBackup(Settings backedUpSettings) {
		return SettingsComparator.isEqual(getCurrentSettings(), backedUpSettings);
	}

}
