package pro.litvinovg.docsettings.settings;

import static pro.litvinovg.docsettings.settings.StorageUtils.*;

import java.io.IOException;
import java.time.Instant;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.TimeZone;
import java.util.TreeMap;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import com.sun.star.beans.IllegalTypeException;
import com.sun.star.beans.Property;
import com.sun.star.beans.PropertyExistException;
import com.sun.star.beans.UnknownPropertyException;
import com.sun.star.beans.XPropertyContainer;
import com.sun.star.beans.XPropertySet;
import com.sun.star.beans.XPropertySetInfo;
import com.sun.star.lang.IllegalArgumentException;
import com.sun.star.lang.WrappedTargetException;
import com.sun.star.uno.UnoRuntime;

import pro.litvinovg.docsettings.exceptions.MalformedXmlException;

public class Backups {

	private static final short REMOVABLE_PROPERTY = 128;
	private static final int OO_XML_MAX_LENGTH = 65535;
	private static final Logger logger = LoggerFactory.getLogger(Backups.class);
	private static final String BACKUP_NAME_PREFIX = "SerializedSettings-";
	private TreeMap<Long,Settings> backups = null;

	public Backups() {
		backups = new TreeMap<Long,Settings>();
	}

	public void add(Long unixTimeStamp, Settings backup) {
		backups.put(unixTimeStamp, backup);
	}
	public void clear() {
		backups.clear();
	}
	public boolean isEmpty() {
		return backups.isEmpty();
	}
	
	public void create(Settings currentSettings) {
		backups.put( getUnixTime(), currentSettings);
	}
	
	public Settings getBackupByTimeStamp(Long timeStamp) {
		return backups.get(timeStamp);
	}
	
	public Settings getLatest() {
		return backups.lastEntry().getValue();
	}
	
	public Set<Long> getTimeStamps() {
		return backups.keySet();
	}

	public void readBackupsFromMetadata(XPropertyContainer userDefinedProps) {
		clear();
		TreeMap<Long, Map<Long, String>> backupMap = getBackupMap(userDefinedProps);
		Map<Long,String> serializedBackups = combineBackupStrings(backupMap);
		for (Entry<Long, String> entry : serializedBackups.entrySet()) {
			Long unixTimeStamp = entry.getKey();
			String backupString = entry.getValue();
			if (!backupString.isEmpty()) {
				backupString = decodeBase64(backupString);
				readBackup(unixTimeStamp, backupString);
			}
		}
	}

	public void writeBackupsToMetadata(XPropertyContainer userDefinedProps) {
		cleanBackupMetadata(userDefinedProps);
		for (Entry<Long, Settings> backup : backups.entrySet()) {
			try {
				addBackupEntries( userDefinedProps, backup);
			} catch (MalformedXmlException | ParserConfigurationException | TransformerException e) {
				logger.debug("Ops", e);
			}
		}
	}
	
	private void cleanBackupMetadata(XPropertyContainer userDefinedProps) {
		XPropertySet propertySet = UnoRuntime.queryInterface(XPropertySet.class, userDefinedProps);
		XPropertySetInfo propSetInfo = propertySet.getPropertySetInfo();
		Property[] props = propSetInfo.getProperties();
		for (int i = 0; i < props.length; i++) {
			try {
				String name = props[i].Name;
				if (name.startsWith(BACKUP_NAME_PREFIX)) {
					userDefinedProps.removeProperty(name);
				}
			} catch (Exception e) {
				logger.debug("Ops", e);
			}
		}		
	}

	private void addBackupEntries(XPropertyContainer userDefinedProps, Entry<Long, Settings> backup) throws MalformedXmlException, ParserConfigurationException, TransformerException {
		String backupString = serializeSettingsXML(backup.getValue());
		backupString = encodeBase64(backupString);
		List<String> parts = splitByOOMaxLength(backupString);
		for (int i = 0; i < parts.size();i++) {
			String part = parts.get(i);
			String name = formatBackupMetaName(backup.getKey(),i); 
			try {
				userDefinedProps.addProperty(name, REMOVABLE_PROPERTY, part);
			} catch (IllegalArgumentException e) {
				logger.debug("Ops", e);
			} catch (PropertyExistException e) {
				//it is ok
				logger.debug("Ops", e);
			} catch (IllegalTypeException e) {
				logger.debug("Ops", e);
			}
			
		}
	}
	
	private void readBackup(Long unixTimeStamp, String backupString) {
		Document backupDoc;
		try {
			backupDoc = getDocFromString(backupString);
			Element configSettings = (Element) getSettingsElement(backupDoc);
			NodeList settingNodes = configSettings.getChildNodes();
			Settings backup = readSettingsFromNodeList(settingNodes);
			add(unixTimeStamp,backup);
		} catch (ParserConfigurationException | SAXException | IOException e) {
			logger.debug("Ops", e);
		} catch (Exception e) {
			logger.debug("Ops", e);
		}
	}

	private TreeMap<Long, Map<Long, String>> getBackupMap(XPropertyContainer userDefinedProps) {
		TreeMap<Long, Map<Long, String>> backupMap = new TreeMap<Long,Map<Long,String>>(); 

		XPropertySet propertySet = UnoRuntime.queryInterface(XPropertySet.class, userDefinedProps);
		XPropertySetInfo propSetInfo = propertySet.getPropertySetInfo();
		Property[] props = propSetInfo.getProperties();
		for (int i = 0; i < props.length; i++) {
			try {
				Property prop = props[i];
				String name = prop.Name;
				if (name.startsWith(BACKUP_NAME_PREFIX)) {
					Long timeStamp = Long.parseLong(getTimeStampFromString(name));
					Long order = Long.parseLong(getOrderFromString(name));
					String stringPart = propertySet.getPropertyValue(prop.Name).toString();
					Map<Long, String> backupParts = null;
					if (backupMap.containsKey(timeStamp)) {
						backupParts = backupMap.get(timeStamp);
					} else {
						backupParts = new TreeMap<Long,String>();
						backupMap.put(timeStamp, backupParts);
					}
					backupParts.put(order, stringPart);
				}
				//logger.debug("Prop name "+ prop.Name + " Prop value " + propertySet.getPropertyValue(prop.Name));
			} catch (UnknownPropertyException | WrappedTargetException e) {
				logger.debug("Ops", e);
			}
		}
		return backupMap;
	}
	
	public void readBackupsFromXML(String metaContent) {
		clear();
		try {
			Document doc = getDocFromString(metaContent);
			Map<Long,String> serializedBackups = getSerializedBackups(doc);
			for (Entry<Long, String> entry : serializedBackups.entrySet()) {
				Long unixTimeStamp = entry.getKey();
				String backupString = entry.getValue();
				if (!backupString.isEmpty()) {
					backupString = decodeBase64(backupString);
					Document backupDoc = getDocFromString(backupString);
					Element configSettings = (Element) getSettingsElement(backupDoc);
					NodeList settingNodes = configSettings.getChildNodes();
					Settings backup = readSettingsFromNodeList(settingNodes);
					add(unixTimeStamp,backup);
				}
			}
		} catch (Exception e) {
			logger.debug(e.getMessage());
			logger.debug("Ops!",e);
		}
	}

	private Settings readSettingsFromNodeList(NodeList settingNodes) throws MalformedXmlException {
		Settings settings = new Settings();
		for (int i = 0; i < settingNodes.getLength(); i++) {
			Node settingNode = settingNodes.item(i);
			ODFSetting setting = SettingFactory.create(settingNode);
			settings.add(setting);
		}
		return settings;
	}

	public String getNewMetaContent(String metaContent) {
		String cleanedMeta = getCleanedMetaContent(metaContent);
		String newMetaContent = "";
		try {
			Document doc = getDocFromString(cleanedMeta);
			for (Entry<Long, Settings> backup : backups.entrySet()) {
				addXMLBackupEntries( doc, backup.getKey(), backup.getValue());
			}
			newMetaContent = serializeDoc(doc);
		} catch (Exception e) {
			logger.debug(e.getMessage());
			logger.debug("Ops!",e);
		}
		return newMetaContent;
	}
	
	
	
	private void addXMLBackupEntries(Document doc, Long timeStamp, Settings settings) throws MalformedXmlException, ParserConfigurationException, TransformerException {
		String backupString = serializeSettingsXML(settings);
		backupString = encodeBase64(backupString);
		List<String> parts = splitByOOMaxLength(backupString);
		for (int i = 0; i < parts.size();i++) {
			String part = parts.get(i);
			addMetadataEntry(doc,timeStamp,part, i);
		}
	}

	private void addMetadataEntry(Document doc, Long timeStamp, String string, int i) throws MalformedXmlException {
		Element element = doc.createElement("meta:user-defined");
		String name = formatBackupMetaName(timeStamp,i); 
		element.setAttribute("meta:name", name);		
		element.setTextContent(string);
		Node metaRoot = getMetaRootElement(doc);
		metaRoot.appendChild(element);
	}

	private String formatBackupMetaName(Long timeStamp, int i) {
		String result = BACKUP_NAME_PREFIX + timeStamp.toString() + "-" + i;
		return result;
	}

	private List<String> splitByOOMaxLength(String text) {
		List<String> array = new ArrayList<String>();
		int index = 0;
		while (index < text.length()) {
			array.add(text.substring(index, Math.min(index + OO_XML_MAX_LENGTH, text.length())));
			index += OO_XML_MAX_LENGTH;
		}
		return array;
	}

	private String serializeSettingsXML(Settings settings) throws ParserConfigurationException, TransformerException {
		Document doc = createNewDoc();
		Element root = doc.createElementNS("urn:oasis:names:tc:opendocument:xmlns:config:1.0","config:config-item-set");
		root.setAttribute("xmlns:config", "urn:oasis:names:tc:opendocument:xmlns:config:1.0");
		root.setAttribute("xmlns:ooo", "http://openoffice.org/2004/office");
		root.setAttribute("config:name", "ooo:configuration-settings");
		doc.appendChild(root);
		appendSettingNodesToDoc(settings, doc, root);
		String result = serializeDoc(doc);
		return result;
	}


	private String getCleanedMetaContent(String metaContent) {
		String cleanedMetaContent = "";
		try {
			Document doc = getDocFromString(metaContent);
			NodeList nodes = getBackupEntries(doc);
			removeNodes(nodes);
			cleanedMetaContent = serializeDoc(doc);
		} catch (Exception e) {
			logger.debug(e.getMessage());
			logger.debug("Ops!",e);
		}
		return cleanedMetaContent;
	}



	private Map<Long,String> getSerializedBackups(Document doc) throws MalformedXmlException {
		NodeList nodes = getBackupEntries(doc);
		TreeMap<Long, Map<Long, String>> backupStringMap = sortBackupStrings(nodes);
		Map<Long,String> result = combineBackupStrings(backupStringMap);
		return result;
	}

	private NodeList getBackupEntries(Document doc) throws MalformedXmlException {
		XPathFactory xPathfactory = XPathFactory.newInstance();
		XPath xpath = xPathfactory.newXPath();
		XPathExpression expr;
		NodeList nodes;
		try {
			expr = xpath.compile("//*[local-name() ='user-defined'][@*[local-name() ='name' and starts-with(.,'"+ BACKUP_NAME_PREFIX + "')]]");
			nodes = (NodeList) expr.evaluate(doc, XPathConstants.NODESET);
		} catch (XPathExpressionException e) {
			logger.debug(e.getLocalizedMessage());
			throw new MalformedXmlException("Xpath expression failed while looking for backup entries");
		}
		return nodes;
	}
	
	private Map<Long, String> combineBackupStrings(TreeMap<Long, Map<Long, String>> backupMap) {
		TreeMap<Long, String> result = new TreeMap<Long, String>();
		for(Map.Entry<Long, Map<Long,String>> entry: backupMap.entrySet()) {
			Long timeStamp = entry.getKey();
			Map<Long, String> parts = entry.getValue(); 
			String combinedString = "";
			for(Map.Entry<Long, String> part: parts.entrySet()) {
				String serlializedString = part.getValue(); 
				combinedString += serlializedString;
			}
			result.put(timeStamp, combinedString);
		}
		return result;
	}
	
	private TreeMap<Long, Map<Long, String>> sortBackupStrings(NodeList nodes) throws MalformedXmlException {
		TreeMap<Long, Map<Long, String>> backupMap = new TreeMap<Long,Map<Long,String>>(); 
		for (int i = 0; i < nodes.getLength(); i++) {
			Node node = nodes.item(i);
			Long timeStamp = getTimeStampFromNode(node);
			Long order = getOrderFromNode(node);
			String stringPart = node.getTextContent();
			Map<Long, String> backupParts = null;
			if (backupMap.containsKey(timeStamp)) {
				backupParts = backupMap.get(timeStamp);
			} else {
				backupParts = new TreeMap<Long,String>();
				backupMap.put(timeStamp, backupParts);
			}
			backupParts.put(order, stringPart);
		}
		return backupMap;
	}	

	private Long getTimeStampFromNode(Node node) throws MalformedXmlException {
		String name = node.getAttributes().getNamedItem("meta:name").getNodeValue();
		String ts = getTimeStampFromString(name);
		Long timeStamp = null;
		try {
			timeStamp = Long.parseLong(ts);
		} catch (NumberFormatException e) {
			throw new MalformedXmlException("timestamp not found for metadata backup entry " + name);
		}
		return timeStamp;
	}

	private String getTimeStampFromString(String name) {
		return name.replaceAll(BACKUP_NAME_PREFIX, "").replaceAll("-[0-9]$","");
	}
	
	private Long getOrderFromNode(Node node) throws MalformedXmlException {
		String name = node.getAttributes().getNamedItem("meta:name").getNodeValue();
		String or = getOrderFromString(name);
		Long order;
		try {
		order = Long.parseLong(or);
		} catch (NumberFormatException e) {
			throw new MalformedXmlException("order not found for metadata backup entry " + name);
		}
		return order;
	}

	private String getOrderFromString(String name) {
		return name.replaceAll(BACKUP_NAME_PREFIX + "[0-9]+-", "");
	}
	
	private long getUnixTime() {
		return System.currentTimeMillis() / 1000L;
	}
	
	public String[] getBackupTimeLabels() {
		String[] timeLabels = new String[backups.size()];
		int i = 0;
		for(Long backupTime: backups.keySet()) {
			timeLabels[i] = getDateTimeLabelsFromUnix(backupTime);
			i++;
		}
		return timeLabels;
	}
	
	private String getDateTimeLabelsFromUnix(Long timeStamp){
		ZoneId zoneId = TimeZone.getDefault().toZoneId();
		DateTimeFormatter formatter = DateTimeFormatter.ofPattern("dd.MM.yyyy HH:mm:ss");
		final Instant instant = Instant.ofEpochSecond(timeStamp);
		LocalDateTime time = LocalDateTime.ofInstant(instant, zoneId); 
    return formatter.format(time);
	}
	
	private Long getUnixFromDateTimeLabels(String formattedTime){
		ZoneId zoneId = TimeZone.getDefault().toZoneId();
    DateTimeFormatter parseFormat = DateTimeFormatter.ofPattern("dd.MM.yyyy HH:mm:ss");
    LocalDateTime time = LocalDateTime.parse(formattedTime, parseFormat); 
    return time.atZone(zoneId).toEpochSecond();		
	}

	public void removeByDateTimeLabel(String dateTimeLabel) {
		if (dateTimeLabel == null || dateTimeLabel.isEmpty()) {
			return;
		}
		Long unix = getUnixFromDateTimeLabels(dateTimeLabel);
		backups.remove(unix);
	}
	
	public Settings getByDateTimeLabel(String dateTimeLabel) throws Exception {
		if (dateTimeLabel == null || dateTimeLabel.isEmpty()) {
			throw new Exception("date time not found");
		}
		Long unix = getUnixFromDateTimeLabels(dateTimeLabel);
		Settings result = backups.get(unix);
		if (result == null) {
			throw new Exception("backup not found");
		}
		return result;
	}


}
